#!/usr/bin/env python
# coding: utf-8

# In[1]:


import os 
import glob
import time
from datetime import datetime

import numpy as np
import matplotlib.pyplot as plt
import pandas as pd

from platform import python_version
print('Python version:',python_version())

# MF-related libraries
import tensorflow as tf
print('Tensor Flow version: ',tf.__version__)
from tensorflow.keras.callbacks import TensorBoard,EarlyStopping,ReduceLROnPlateau
from tensorflow.keras.models import Sequential

from sklearn.preprocessing import MinMaxScaler


# # Weather Forecasting - Machine Learning (RNN-LSTM) 

# Using hourly weather data (temperature, pressure and wind speed) from US cities and the machine learning algortithm for weather forecasting. Long short-term memory (LSTM) is an artificial recurrent neural network (RNN) architecture used in the field of deep learning and can be also applied in the case of time series analysis. We run this example on a CPU.

# ## Prepare the dataset

# In[2]:


#Source: https://www.kaggle.com/selfishgene/historical-hourly-weather-data

# directory with data files in the csv format
inputDir="../data/us-hourly-weather-data"

# I would like to forceast data for that city:
target_city="Boston"

# Cities from which data should be used in the model
input_cities=[target_city, "Chicago", "Toronto","Detroit","Indianapolis","Charlotte"] #,
            #  "New York","Portland","Montreal","Pittsburgh"]

# Taget parameters to forecast
target_params=["temperature","pressure","wind_speed"] #,"humidity"] #,"wind_direction"]
input_params=target_params

# whether to add additional and (hopefully) usefull information
# If true adds doy dependence
add_doy=True

# load / stack data
data_size=0
data_list=[]
feature_list=[]
feature_dict=dict()
params_minmax_dict=dict()

# small data correction algorithm for the input data
params_minmax_dict["pressure"]=list([950,1050]) # cut off values for pressure 
params_minmax_dict["wind_speed"]=list([0,20]) # cut off values for pressure 

# load data and correct some obvious outliers
no_features=0
for csv_path in glob.iglob(inputDir+'/'+'*.csv'):
    feature_i_name=csv_path.split('.csv')[0]
    feature_i_name=feature_i_name.rsplit('/')[3]
    print(feature_i_name,': ',csv_path)
    feature_list.append(feature_i_name)
    # Check and concat data
    df_i = pd.read_csv(csv_path)
    # fix the missing data
    df_i=df_i.fillna(method='bfill')
    df_i=df_i.fillna(method='ffill')
            # correct bad data
    if(feature_i_name in params_minmax_dict.keys()):
        print("Correcting (clipping)", feature_i_name)
        df_i[input_cities].clip(lower=params_minmax_dict[feature_i_name][0])
        df_i[input_cities].clip(upper=params_minmax_dict[feature_i_name][1])
    if (feature_i_name != 'city_attributes'):
        df_i.set_index('datetime')  

    feature_dict[feature_list[no_features]]=df_i
    no_features+=1
 


# In[3]:


print("Adding: ",target_params[0])
data_tmp=feature_dict[target_params[0]]
for i in range(1,len(target_params)):
    print("Adding: ",target_params[i])
    df_i=pd.DataFrame(feature_dict[target_params[i]])
    frames=[data_tmp,df_i]
    data_tmp=pd.concat(frames, axis=1)
data_tmp.index=feature_dict[target_params[0]]['datetime']

data=data_tmp[input_cities]
# additional variables as an input to the ML algorithm
if (add_doy == True):
    data_more=data_tmp.index
    # retrieve doy param and add it to the data set
    data_doy=[datetime.strptime(d, '%Y-%m-%d %H:%M:%S').strftime("%j") for d in data_more]
    data_tmp["doy"]=data_doy 
    input_params=[target_city,"doy"]
    print("Adding:  doy")
    data['doy'] = data_doy #data.assign(e=data_doy)

data_values=data.values


# ## Display dataset

# In[4]:


# Display first 5 rows and all columns of the data set that will be used for ML
data.head(5)


# Plot the created dataset (input to the model).
# There are still some obvious outliers in the data. However, this can be corrected in the future.

# In[5]:


data.plot(subplots=True,figsize=(30,30))


# ## Set up the ML-related parameters

# Here we define the important params for our machine-learning algorithm. This inlcudes the fraction of data to be used for training (TRAIN_SPLIT) and validation (1-TRAIN_SPLIT), batch size (BATCH_SIZE), number of epochs of the MĹ algorithm (EPOCHS) and the number of steps per epoch (STEPS_PER_EPOCH). In addition,one needs to define the layer size (number of neurons per layer) and the number of layers, if one would like to extent the neural network. These are the parameters that one would need to alter when benchmariking/improving our ML algorithm.

# In[6]:


# fraction of data used for training 
TRAIN_SPLIT = 0.9
BATCH_SIZE =64
EPOCHS=5
STEPS_PER_EPOCH=15

# Here we go with the minimization for mean absolute error
OPTIMIZATION_TYPE='mae'

# The array can be used for ML performance tests and the results can be displayed in Tensorboard
#LAYER_SIZES=[64,128,256]
#DENSE_LAYERS=[0,1,2]
# For this particular example, we test with a sinlge LSTM layer with 64 neurons
layer_size=64
dense_layer=0

# We will use a sequence-length of 1344 (instead of the whole period), which means that each random sequence (batch) contains 
#observations for 8 weeks. One time-step corresponds to one hour, so 24 x 7 time-steps corresponds 
#to a week, and 24 x 7 x 8 corresponds to 8 weeks.
SEQUENCE_LEN = 24 * 7 * 8


# ### Prepare target data for learning
# The following is the number of time-steps that we will shift the target-data. Our data-set is resampled to have an observation for each hour, so there are 24 observations for 24 hours.
# If we want to predict the weather 24 hours into the future, we shift the data 24 time-steps. If we want to predict the weather n days into the future, we shift the data n * 24 time-steps.

# 

# In[7]:


shift_days = 1
shift_steps = shift_days * 24  # Number of hours.

# Create a new data-frame with the time-shifted data.
data_target=data[target_city].shift(-shift_steps)


# The shifted data-frame is confusing because Pandas keeps the 
# original time-stamps even though we have shifted the data.

# In[8]:


data.head(shift_steps+5)


# In[9]:


data_target.head(5)


# In[10]:


# The last observations are NaN, because the data has been shifted backwards
data_target.tail()


# In[11]:


#These are the input-signals:
x_data= data_values[:-shift_steps]
print('X Data shape:',x_data.shape)
#These are the output-signals (or target-signals)
y_data = data_target.values[:-shift_steps]

# Data shape: number of epoch x number of target parameters (temp., pressure, wind_speed)
print('Y Data shape:',y_data.shape)
num_data = len(x_data)
print('############################')
print('Number of epochs:',num_data)
print('Number of input params:',x_data.shape[1])
print('Number of output params:',y_data.shape[1])
print('############################')


# ## Prepare training and validation datasets
# Based on the defined TRAIN_SPLIT, we prepare the two data sets: one used for training our model, and the other for validation purposes. The latter will not be "seen" by the ML algorithm.

# In[12]:


num_train = int(TRAIN_SPLIT * num_data)
num_test=num_data-num_train

if (len(input_params)>1):
    num_x_signals = x_data.shape[1]
else:
    num_x_signals = x_data.shape[0]
#number of output signals
if (len(target_params)>1):
    num_y_signals = y_data.shape[1]
else:
    num_y_signals = y_data.shape[0]
print("Num of epochs (all):",num_data,"Num of epoch for training:", num_train,"Num of epochs for validation:",num_test)


# In[13]:


x_train = x_data[0:num_train]
x_test = x_data[num_train:]

y_train = y_data[0:num_train]
y_test = y_data[num_train:]

if (len(input_params)==1):
    x_train =x_train.reshape(-1,1)
    x_test =x_test.reshape(-1,1)
if (len(target_params)==1):
    y_train =y_train.reshape(-1,1)
    y_test =y_test.reshape(-1,1)

print(x_train,y_train)


# ### Rescale data sets
# ML works better on the scaled data sets (default 0,1) for MinMaxScaler()

# In[14]:


#scale (training) data 
x_scaler = MinMaxScaler()
#Compute the minimum and maximum to be used for later scaling.
x_train_scaled = x_scaler.fit_transform(x_train)
#We use the same scaler-object for the input-signals in the test-set.
x_test_scaled=x_scaler.transform(x_test)
# create a separate scaler object for y data (in this case one could use the same in princple as for x)
y_scaler = MinMaxScaler()
y_train_scaled = y_scaler.fit_transform(y_train)
y_test_scaled = y_scaler.transform(y_test)

print(x_train_scaled,y_train_scaled)


# ## Generate Batch Data
# 
# Instead of training the RNN on the complete sequences of observations, use the function to create a batch of shorter sub-sequences picked at random from the training-data

# In[15]:


def batch_generator(batch_size, sequence_length):
    """
    Generator function for creating random batches of training-data.
    """

    # Infinite loop.
    while True:
        # Allocate a new array for the batch of input-signals.
        x_shape = (batch_size, sequence_length, num_x_signals)
        x_batch = np.zeros(shape=x_shape, dtype=np.float16)

        # Allocate a new array for the batch of output-signals.
        y_shape = (batch_size, sequence_length, num_y_signals)
        y_batch = np.zeros(shape=y_shape, dtype=np.float16)

        # Fill the batch with random sequences of data.
        for i in range(batch_size):
            # Get a random start-index.
            # This points somewhere into the training-data.
            idx = np.random.randint(num_train - sequence_length)
            
            # Copy the sequences of data starting at this index.
            x_batch[i] = x_train_scaled[idx:idx+sequence_length]
            y_batch[i] = y_train_scaled[idx:idx+sequence_length]
        
        yield (x_batch, y_batch)


# In[16]:


# create BATCH_SIZE batches of random data of SEQUENCE_LEN len.
generator = batch_generator(batch_size=BATCH_SIZE,
                            sequence_length=SEQUENCE_LEN)


# In[17]:


x_batch, y_batch = next(generator)
# check the no of sequences for LSTM
print(x_batch.shape)
print(y_batch.shape)


# In[18]:


# plot the (first) sequence
batch = 0   # First sequence in the batch.
signal = 0  # First signal (parameter) from the n input-signals.
seq = x_batch[batch, :, signal]
print("X_BATCH: Batch:",batch,"Signal:",signal)
plt.plot(seq)


# In[19]:


# plot also the output signal
seq = y_batch[batch, :, signal]
#print(y_batch)
print("Y_BATCH: Batch:",batch,"Signal:",signal)
plt.plot(seq)


# ### Validation data set
# 
# There is always a risk of overfitting the model to the training set so it does not generalize well to unseen data. One can monitor the model's performance utilizing the test set after each epoch.
# 
# The batch generator randomly selects a batch of short sequences from the training-data and uses that during training. In the csse of validation data, one instead runs through the entire data from the test set and measure the prediction accuracy on that entire sequence.

# In[20]:


validation_data = (np.expand_dims(x_test_scaled, axis=0),
                   np.expand_dims(y_test_scaled, axis=0))


# # Create & Train the Recurrent Neural Network - LSTM

# The main part of the ML algorithm. One defines here the ML algortithm type, its layers, etc.

# In[21]:


# Define the model as sequential
model = Sequential()

# Add LSTM layer
model.add(tf.keras.layers.LSTM(layer_size, 
                               return_sequences=True,
                               input_shape=(None, num_x_signals,),activation='relu'))
# Add dense layers in case dense_layes is not zero
for dense_layer_i in range(0,dense_layer):
    model.add(tf.keras.layers.Dense(layer_size/(2*(dense_layer_i+1)),activation='relu'))
    model.add(tf.keras.layers.Dropout(0.2))
model.add(tf.keras.layers.Dense(num_y_signals)) # as we are interested in n features

# choose optimization type
optimizer = tf.keras.optimizers.RMSprop(lr=1e-3)
model.compile(optimizer=optimizer, loss=OPTIMIZATION_TYPE,metrics=['accuracy'])
# display summary of the model
model.summary()


# ### Define calback functions to use
# Calback functions are additional tools helping to improve the performance of our ML algorithm and are very usefull at the diagonstic stage.
# 
# - Tensorboard: used for benchmarking optimization of the machine learning algorithm
# - EarlyStopping: callback for stopping the optimization when performance worsens on the validation-set
# - ReduceLROnPlateau: reduces the learning-rate for the optimizer if the validation-loss has not improved 
# since the last epoch (as indicated by patience=0). The learning-rate will be reduced by multiplying 
# it with the given factor. We set a start learning-rate of 1e-3, so multiplying it by 0.1 gives 
#  a learning-rate of 1e-4. We don't want the learning-rate to go any lower than this.
# 
# Learning rate: gradient descent algorithms multiply the gradient by a scalar known as the learning rate (also sometimes called step size) to determine the next point. 
# If the gradient magnitude is 2.5 and the learning rate is 0.01, then the gradient descent algorithm will pick the next point 0.025 away from the previous point.

# In[22]:


# logging for optimizing
# Tensorboard 
NAME= "METEO-{}-nodes-{}-dense-{}".format(layer_size,dense_layer,int(time.time()))
callback_tensorboard = TensorBoard(log_dir='logdir/{}'.format(NAME))

# Early Stopping
callback_early_stopping = EarlyStopping(monitor='val_loss',
                                        patience=3, verbose=1)

#ReduceLROnPlateau
callback_reduce_lr = ReduceLROnPlateau(monitor='val_loss',
                                       factor=0.1,
                                       min_lr=1e-4,
                                       patience=0,
                                       verbose=1)

callbacks = [callback_early_stopping, callback_tensorboard, callback_reduce_lr]


# ## Training/fitting the model

# In[ ]:


print("MODEL: ",NAME)
model.fit(x=generator,
          epochs=EPOCHS,
          steps_per_epoch=STEPS_PER_EPOCH,
          validation_data=validation_data,callbacks=callbacks)


# # Performance Tests 
# 
# Once we obtained our model, we can validate its performance using traning and validation (test) data sets. 
# Evaluation can be performed using various metrics, such as mean squared error (mse) or mean absolute error (mae)

# In[ ]:


# Evaluate the model on the test data using `evaluate`

result = model.evaluate(x=np.expand_dims(x_test_scaled, axis=0),
                        y=np.expand_dims(y_test_scaled, axis=0))
print("loss (test set): %8.4f " % result[0])


# ## ML model vs training data and test data

# In[ ]:


# helper function to plot all the data sets

def plot_comparison(start_idx, length=100, train=True):
    """
    Plot the predicted and true output-signals.
    
    :param start_idx: Start-index for the time-series.
    :param length: Sequence-length to process and plot.
    :param train: Boolean whether to use training- or test-set.
    """
    
    if train:
        # Use training-data.
        x = x_train_scaled
        y_true = y_train
    else:
        # Use test-data.
        x = x_test_scaled
        y_true = y_test
    
    # End-index for the sequences.
    end_idx = start_idx + length
    
    # Select the sequences from the given start-index and
    # of the given length.
    x = x[start_idx:end_idx]
    y_true = y_true[start_idx:end_idx]
    
    # Input-signals for the model.
    x = np.expand_dims(x, axis=0)

    # Use the model to predict the output-signals.
    y_pred = model.predict(x)
    
    # The output of the model is between 0 and 1.
    # Do an inverse map to get it back to the scale
    # of the original data-set.
    y_pred_rescaled = y_scaler.inverse_transform(y_pred[0])
    
    # For each output-signal.
    for signal in range(len(target_params)):
        # Get the output-signal predicted by the model.
        print('Signal: ',signal)
        
        signal_pred = y_pred_rescaled[:, signal]
        
        # Get the true output-signal from the data-set.
        signal_true = y_true[:, signal]

        # calculate some statistics
        mae=tf.keras.metrics.mean_absolute_error(signal_true,signal_pred).numpy()
        mape=np.mean(np.abs((signal_true - signal_pred) / signal_true)) * 100
        rms=np.sqrt(tf.keras.metrics.mean_squared_error(signal_true,signal_pred).numpy())
        rmspre=np.sqrt(np.mean((signal_true - signal_pred)/signal_true)**2)*100
        #prmse=np.sqrt(np.mean(signal_true-signal_pred)**2/np.sum(signal_true**2))*100
        txt_stats="RMS: %4.2f (%4.2f %%) MAE: %4.2f (%4.2f%%)" % (rms,rmspre,mae,mape)

        fig, ax = plt.subplots(figsize=(35,20))
        # Plot and compare the two signals.
        plt.plot(signal_true, label='true')
        plt.plot(signal_pred, label='pred',linewidth=2)
        
        ax.tick_params(axis='both', which='major', labelsize=24)
        ax.tick_params(axis='both', which='minor', labelsize=24)
        
        # Plot labels etc.
        ax.text(0.05, 0.95, txt_stats, transform=ax.transAxes, fontsize=28,
        verticalalignment='top')
        plt.ylabel(target_params[signal],fontsize=24)
        plt.legend(fontsize=24,loc='upper right')
        plt.show()


# ### Model vs the train set
# Here one compares the performance of the model on the data that were used for generating (training) that model.

# In[ ]:


# plot 
# plot true vs predicted on the training set
plot_comparison(start_idx=0, length=num_train, train=True)


# ### Model vs validation (test) set 
# 
# Our ML algorithm did not see those data before and we verify the performance of our model on the recent (1-VAL_SPLIT) fraction of data. 

# In[ ]:


# plot true vs predicted on the test set
plot_comparison(start_idx=0, length=num_train, train=False)


# # Final Remarks
# 
# One should distinguish between *causation* and *correlation* in the data. 
# The neural network can easily discover patterns in the data that are only temporary correlated. This leads to the problem of lack of generalization, i.e., our algorithm will not perform well on the unseen data. One should select input- and output-data where a *causal* relationship probably exists, e.g., day of year and meteo parameters.
# 
# In order to enhance the performance of our ML alorithm one would need to investigate/alter the parameters defining our neural network, such as layer_size, number of dense layers, batch_size, number of epochs, etc. One can also iterate over the same data set and choose the best model using "ModelCheckpoint".

# In[ ]:




